#include <Python.h>
#define crypto_MODULE
#include "crypto.h"

static char *CVSid = "@(#) $Id$";

static void crypto_CRL_dealloc(crypto_CRLObj *self);

static const char *crl_reasons[] = {
        /* CRL reason strings */
        "unspecified",
        "keyCompromise",
        "CACompromise",
        "affiliationChanged",
        "superseded",
        "cessationOfOperation",
        "certificateHold",
        "removeFromCRL",
        /* Additional pseudo reasons */
        "holdInstruction",
        "keyTime",
        "CAkeyTime"
};

#define NUM_REASONS (sizeof(crl_reasons) / sizeof(char *))

int unpack_revinfo(ASN1_TIME **prevtm, int *preason, ASN1_OBJECT **phold, ASN1_GENERALIZEDTIME **pinvtm, const char *str)
{
        char *tmp = NULL;
        char *rtime_str, *reason_str = NULL, *arg_str = NULL, *p;
        int reason_code = -1;
        int ret = 0;
        unsigned int i;
        ASN1_OBJECT *hold = NULL;
        ASN1_GENERALIZEDTIME *comp_time = NULL;
        tmp = BUF_strdup(str);

        p = strchr(tmp, ',');
        rtime_str = tmp;
        if (p) {
		*p = '\0';
		p++;
		reason_str = p;
		p = strchr(p, ',');
		if (p) {
			*p = '\0';
			arg_str = p + 1;
		}
	}	
        if (prevtm) {
		*prevtm = ASN1_UTCTIME_new();
		if (!ASN1_UTCTIME_set_string(*prevtm, rtime_str))
			goto err;
	}
        if (reason_str) {
		for (i = 0; i < NUM_REASONS; i++) {
			if(!strcasecmp(reason_str, crl_reasons[i])) {
				reason_code = i;
				break;
			}
		}
		if (reason_code == OCSP_REVOKED_STATUS_NOSTATUS)
			goto err;

		if (reason_code == 7)
			reason_code = OCSP_REVOKED_STATUS_REMOVEFROMCRL;
		else if (reason_code == 8) {
			if (!arg_str)
				goto err;
			reason_code = OCSP_REVOKED_STATUS_CERTIFICATEHOLD;
			hold = OBJ_txt2obj(arg_str, 0);

			if (!hold)
				goto err;
			if (phold) *phold = hold;
		}
		else if ((reason_code == 9) || (reason_code == 10)) {
			if (!arg_str)
				goto err;
			comp_time = ASN1_GENERALIZEDTIME_new();
			if (!ASN1_GENERALIZEDTIME_set_string(comp_time, arg_str))
				goto err;
			if (reason_code == 9)
				reason_code = OCSP_REVOKED_STATUS_KEYCOMPROMISE;
			else
				reason_code = OCSP_REVOKED_STATUS_CACOMPROMISE;
		}
	}

        if (preason) *preason = reason_code;
        if (pinvtm) *pinvtm = comp_time;
        else ASN1_GENERALIZEDTIME_free(comp_time);
	
        ret = 1;

 err:
        if (tmp) OPENSSL_free(tmp);
        if (!phold) ASN1_OBJECT_free(hold);
        if (!pinvtm) ASN1_GENERALIZEDTIME_free(comp_time);

        return ret;
}
int make_revoked(X509_REVOKED *rev, const char *str)
{
        char *tmp = NULL;
        int reason_code = -1;
        int i, ret = 0;
        ASN1_OBJECT *hold = NULL;
        ASN1_GENERALIZEDTIME *comp_time = NULL;
        ASN1_ENUMERATED *rtmp = NULL;

        ASN1_TIME *revDate = NULL;

        i = unpack_revinfo(&revDate, &reason_code, &hold, &comp_time, str);

        if (i == 0)
                goto err;

        if (rev && !X509_REVOKED_set_revocationDate(rev, revDate))
                goto err;

        if (rev && (reason_code != OCSP_REVOKED_STATUS_NOSTATUS)) {
		rtmp = ASN1_ENUMERATED_new();
		if (!rtmp || !ASN1_ENUMERATED_set(rtmp, reason_code))
			goto err;
		if (!X509_REVOKED_add1_ext_i2d(rev, NID_crl_reason, rtmp, 0, 0))
			goto err;
	}
	
        if (rev && comp_time) {
		if (!X509_REVOKED_add1_ext_i2d(rev, NID_invalidity_date, comp_time, 0, 0))
			goto err;
	}
        if (rev && hold) {
		if (!X509_REVOKED_add1_ext_i2d(rev, NID_hold_instruction_code, hold, 0, 0))
			goto err;
	}

        if (reason_code != OCSP_REVOKED_STATUS_NOSTATUS)
                ret = 2;
        else ret = 1;
 err:
        if (tmp) OPENSSL_free(tmp);
        ASN1_OBJECT_free(hold);
        ASN1_GENERALIZEDTIME_free(comp_time);
        ASN1_ENUMERATED_free(rtmp);
        ASN1_TIME_free(revDate);

        return ret;
}

static char crypto_CRL_make_revoked_doc[] = "";
static PyObject *
crypto_CRL_make_revoked(crypto_CRLObj *self, PyObject *args)
{
    char *t, *id;
    X509_REVOKED *r;
    BIGNUM *serial=NULL;
    ASN1_INTEGER *tmpser;

    if (!PyArg_ParseTuple(args, "ss:make_revoked", &t, &id))
        return NULL;

    r = X509_REVOKED_new();
    make_revoked(r, t);
    if (!BN_hex2bn(&serial, id))
	    return 0;
    tmpser = BN_to_ASN1_INTEGER(serial, NULL);
    BN_free(serial);
    serial = NULL;
    X509_REVOKED_set_serialNumber(r, tmpser);
    ASN1_INTEGER_free(tmpser);

    X509_CRL_add0_revoked(self->crl,r);

    Py_INCREF(Py_None);
    return Py_None;
}

crypto_CRLObj *
crypto_CRL_New(X509_CRL *crl)
{
    crypto_CRLObj *self;

    self = PyObject_New(crypto_CRLObj, &crypto_CRL_Type);
    self->crl = crl;
    return self;
}


/*
 * ADD_METHOD(name) expands to a correct PyMethodDef declaration
 *   {  'name', (PyCFunction)crypto_PKCS12_name, METH_VARARGS, crypto_PKCS12_name_doc }
 * for convenience
 */
#define ADD_METHOD(name)        \
    { #name, (PyCFunction)crypto_CRL_##name, METH_VARARGS, crypto_CRL_##name##_doc }
static PyMethodDef crypto_CRL_methods[] =
{
    ADD_METHOD(make_revoked),
    { NULL, NULL }
};
#undef ADD_METHOD


static PyObject *
crypto_CRL_getattr(crypto_CRLObj *self, char *name)
{
    return Py_FindMethod(crypto_CRL_methods, (PyObject *)self, name);
}

static void
crypto_CRL_dealloc(crypto_CRLObj *self)
{
    PyObject_Del(self);
}

PyTypeObject crypto_CRL_Type = {
    PyObject_HEAD_INIT(NULL)
    0,
    "CRL",
    sizeof(crypto_CRLObj),
    0,
    (destructor)crypto_CRL_dealloc,
    NULL, /* print */
    (getattrfunc)crypto_CRL_getattr,
};

int
init_crypto_crl(PyObject *dict)
{
    crypto_CRL_Type.ob_type = &PyType_Type;
    Py_INCREF(&crypto_CRL_Type);
    PyDict_SetItemString(dict, "CRLType", (PyObject *)&crypto_CRL_Type);
    return 1;
}

